/* pci.c */

#define __KERNEL__

#include <asm/io.h>
#include <vgakeyboard.h>
#include "test.h"

#define SCREEN_ADR 0xb8000

static unsigned char pci_read_config_byte(int device, int where) {
        outl(0x80000000 | ((device & 0xffff) << 8) | where, 0xCF8);
        return inb(0xCFC | (where&3));
}

static void pci_write_config_byte(int device, int where, unsigned char value) {
        outl(0x80000000 | ((device & 0xffff) << 8) | where, 0xCF8);
        outb(value, 0xCFC | (where&3));
}

static void modify_value(int select);

extern struct vars * const v;
unsigned short *scr=(short *)SCREEN_ADR;

#include "screen_buffer.h"

/* turn off cursor */
static void cursor_off()
{
  int i;

  outb(0xb, 0x3d4);
  i=inb(0x3d5);
  outb(0xb, 0x3d4);
  outb(i & ~0x1f, 0x3d5);
}

/* turn on cursor */
static void cursor_on()
{
  int i;

  outb(0xb, 0x3d4);
  i=inb(0x3d5);
  outb(0xb, 0x3d4);
  outb((i & ~0x1f)|14, 0x3d5);
}

/* set cursor pos */
static void cursor_pos(int x, int y)
{
  int i=y*80+x;

  outb(0xe, 0x3d4);
  outb(i >> 8, 0x3d5);
  outb(0xf, 0x3d4);
  outb(i, 0x3d5);
}

/* clear screen */
void clear_screen()
{
  int i;

  cursor_off();
  for(i=0;i < 80*25;i++)
    scr[i]=0x0720;
}

static char hex[16]="0123456789ABCDEF";

static int inttoascii(char *s, unsigned char num)
{
  char *t=s;

  if(num > 9) {
    if(num > 99)
      *t++='0'+num/100;
    *t++='0'+num/10%10;
  }
  *t++='0'+(num % 10);
  return t-s;
}

static void print_hex(int x, int y, unsigned char num, int color)
{
  scr[x+y*80]=(color << 8)|hex[num >> 4];
  scr[x+y*80+1]=(color << 8)|hex[num & 15];
}

static void print_num(int x, int y, unsigned char num, int color)
{
  if(num > 99) {
    scr[x+y*80]=(color << 8)|(num/100)|'0';
    scr[x+y*80+1]=(color << 8)|(num/10%10)|'0';
    scr[x+y*80+2]=(color << 8)|(num%10)|'0';
  } else if(num > 9) {
    scr[x+y*80]=(color << 8)|(num/10)|'0';
    scr[x+y*80+1]=(color << 8)|(num%10)|'0';
    scr[x+y*80+2]=(color << 8)|' ';
  } else {
    scr[x+y*80]=(color << 8)|(num%10)|'0';
    scr[x+y*80+1]=(color << 8)|' ';
    scr[x+y*80+2]=(color << 8)|' ';
  }
}

static void print_binary(int x, int y, unsigned char num, int color)
{
  int i;

  for(i=0;i<8;i++)
    scr[y*80+x+i+(i >= 4)]=(color << 8)|((num & (1 << (7-i)))?'1':'0');
  scr[y*80+x+4]=(color << 8)|' ';
}

static void print_string(int x, int y, char *str, unsigned char color)
{
  unsigned char *s=str;

  while(*s)
    scr[(x++)+y*80]=(color << 8)|(*s++);
}

static void print_pci_map()
{
  int i;

  clear_screen();

  print_string(0, 0, "PCI MAP", 15);

  /* Draw border */
  scr[3*80+2]=(11 << 8)|218;
  for(i=3*80+3;i<3*80+54;i++)
    scr[i]=(11 << 8)|196;
  for(i=4*80+2;i<21*80+2;i+=80)
    scr[i]=(11 << 8)|179;

  /* Draw headers */
  for(i=0;i<16;i++)
    print_hex(4+i*3+((i & 8)?3:0), 2, i, 15);
  for(i=0;i<16;i++)
    print_hex(0, 4+i+((i & 8)?1:0), i << 4, 15);
}

#define coordx(i) (4+(i & 15)*3+((i & 8)?3:0))
#define coordy(i) (4+(i >> 4)+((i & 128)?1:0))

void pci_edit()
{
  static unsigned char select=0, newselect=0;
  unsigned char key, byte;
  int i, init=0, clear=0;

  print_pci_map();

  /* Draw PCI Data */
  for(i=0;i<256;i++)
    print_hex(coordx(i), coordy(i), pci_read_config_byte(0, i), 7);

  print_string(58, 4, "Selected Byte:", 11);
  print_string(58, 6, "Value:  ", 11);

  while(1) {
    if(!init)
      print_string(0, 23,
      "Enter to modify, t) Run Memtest, q) Exit -- Cursor Keys to select ", 3);

    if(newselect != select || !init) {
      /* dehilite old */
      print_hex(coordx(select), coordy(select),
                pci_read_config_byte(0, select), 7);
      select=newselect;

      /* hilite selected */
      byte=pci_read_config_byte(0, select);
      print_hex(coordx(select), coordy(select), byte, 15);

      print_num(66, 7, byte, 15);
      print_binary(66, 6, byte, 15);
      print_hex(73, 4, select, 15);
      init=1;
    }
    
    /* wait for keypress */
    while((key=get_key()) == 0 || (key & 0x80));
    if(clear) {
      print_string(56, 14, "                       ", 7);
      clear=0;
    }

    switch(key) {
      case SCANCODE_CURSORLEFT: newselect--; break;
      case SCANCODE_CURSORDOWN: newselect+=16; break;
      case SCANCODE_CURSORRIGHT: newselect++; break;
      case SCANCODE_CURSORUP: newselect-=16; break;
      case SCANCODE_ESCAPE: case SCANCODE_Q:
        /* tell the BIOS to do a warm start */
        *((unsigned short *)0x472) = 0x1234;
        cache_on();
        outb(0xfe,0x64);
        break;
      case SCANCODE_T:
        init=0;
        v->firsttime=2;
        return;
      case SCANCODE_M: case SCANCODE_ENTER:
        modify_value(select);
        init=0;
        clear=1;
        break;
    }
  }
}

static void modify_value(int select)
{
  enum types { DECIMAL, HEX, BINARY };
  static char type=0, digit=0;
  unsigned char key=0, value=pci_read_config_byte(0, select);
  unsigned char cursor=0, limit[3]={2, 1, 7};
  int val, action=0;
  char buf[24], *s=buf;

  print_string(0, 23,
      "TAB cycles numeric types -- Press Enter to save or ESC to cancel  ", 3);
  print_string(58, 11, "Input:", 11);
  cursor_on();

  if(type == 0)
    cursor=(value == 0)?0:(value < 10)?1:2;
  else
    cursor=0;

  while(key != SCANCODE_ESCAPE && key != SCANCODE_ENTER) {
    print_string(58, 10, (type == 0)?"Enter Decimal Value:":
                         (type == 1)?"Enter Hex Value:    ":
                                     "Enter Binary Value: ", 11);
    if(type == 0)      print_num(66, 11, value, 0x1f);
    else if(type == 1) print_hex(66, 11, value, 0x1f);
    else if(type == 2) print_binary(66, 11, value, 0x1f);

    cursor_pos(66+cursor+(cursor > 3), 11);

    /* wait for keypress */
    digit=-1;
    while((key=get_key()) == 0 || (key & 0x80));
    switch(key) {
    case SCANCODE_1 ... SCANCODE_9: digit=key-SCANCODE_1+1; break;
    case SCANCODE_0: digit=0; break;
    case SCANCODE_A: digit=10; break;
    case SCANCODE_B: digit=11; break;
    case SCANCODE_C: digit=12; break;
    case SCANCODE_D: digit=13; break;
    case SCANCODE_E: digit=14; break;
    case SCANCODE_F: digit=15; break;
    case SCANCODE_TAB:
      print_string(64, 11, "            ", 7);
      type=(type+1) % 3;
      if(type == 0)
        cursor=(value == 0)?0:(value < 10)?1:2;
      else
        cursor=0;
      continue;
    case SCANCODE_ENTER:
      action=1;
      continue;
    case SCANCODE_ESCAPE:
      continue;
    case SCANCODE_BACKSPACE:
      if(type == 0) {
        if(cursor < 1)
          continue;
        if(cursor == 1) {
          if(value > 99) value %= 100;
          else if(value > 9) value %= 10;
          else value=0;
          cursor--;
        } else {
          if(value > 99) value /= 10, cursor++;
          else value /= 10;
          cursor--;
        }
        continue;
      }
      /* flow through */
    case SCANCODE_CURSORLEFT:
      if(cursor > 0)
        cursor--;
      continue;
    case SCANCODE_CURSORRIGHT:
      if(type == 0) {
        if(cursor == 0 && !value)
          continue;
        if(cursor == 1 && value < 10)
          continue;
      }
      if(cursor != limit[(int)type])
        cursor++;
      continue;
    }

    if(digit == -1 || (type == 0 && digit > 9) || (type == 2 && digit > 1))
      continue;

    val=value;

    if(type == 0) {
      if(cursor == 0) {
        if(val < 10) val=digit;
        else if(val < 100) val=(val % 10)+digit*10;
        else val=(val % 100)+digit*100;
      } else if(cursor == 1) {
        if(val < 10) val=val*10+digit;
        else if(val < 100) val=val-(val % 10)+digit;
        else val=val-val%100+val%10+digit*10;
      } else {
        if(val < 100) val=val*10+digit;
        else val=val-val%10+digit;
      }
      if(val > 255)
        continue;
      value=val;
    } else if(type == 1) {
      if(cursor == 0)
        value=(value & 0x0f)|(digit << 4);
      else
        value=(value & 0xf0)|digit;
    } else
      value=(value & ~(1 << (7-cursor)))|(digit << (7-cursor));

    /* move cursor forward */
    if(cursor != limit[(int)type] && (type != 0 || cursor || digit))
      cursor++;
  }

  print_string(58, 10, "                    ", 7);
  print_string(58, 11, "                    ", 7);
  cursor_off();

  if(action) {
    if(pci_read_config_byte(0, select) == value) {
      print_string(56, 14, "   No change made!", 14);
    } else {
      pci_write_config_byte(0, select, value);
      val=pci_read_config_byte(0, select);
      if(val != value) {
        s=buf;
        memcpy(s, "Wrote ", 6);
        s+=6;
        s+=inttoascii(s, value);
        memcpy(s, " but read ", 10);
        s+=10;
        s+=inttoascii(s, val);
        *s++='.';
        *s='\0';
        print_string(56, 14, buf, 12);
      } else
        print_string(56, 14, "  Write Successful!", 10);
    }
  } else
    print_string(56, 14, "   Edit Cancelled!", 14);
}
